@@ -2,17 +2,19 @@ from copy import deepcopy
 from functools import partial
 from typing import Type, Sequence, Tuple, Dict, List, Optional, Callable, Any
 
-from django.http import Http404
+from marshmallow import INCLUDE
+from marshmallow import ValidationError
 from rest_framework.authentication import BaseAuthentication
 from rest_framework.permissions import BasePermission
 from rest_framework.request import Request
-from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT, HTTP_404_NOT_FOUND
+from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT, HTTP_404_NOT_FOUND, \
+    HTTP_400_BAD_REQUEST
 from sqlalchemy.exc import StatementError
 from sqlalchemy.orm import Query
 from sqlalchemy_filters import apply_filters
 from sqlalchemy_filters import apply_pagination
 
-from drf_json_api_utils import json_api_spec_http_methods, JsonApiResourceViewBuilder, plugins
+from drf_json_api_utils import json_api_spec_http_methods, JsonApiResourceViewBuilder, plugins, CustomField
 from drf_json_api_utils.sql_alchemy.constructors import auto_construct_schema, AlchemyRelation
 from drf_json_api_utils.sql_alchemy.types import AlchemyComputedFilter
 from .namespace import _TYPE_TO_SCHEMA
@@ -60,6 +62,7 @@ class AlchemyJsonApiViewBuilder:
         self._before_list_callback = None
         self._after_list_callback = None
         self._relations = []
+        self._custom_fields = {}
         self._computed_filters = {}
         self._skip_plugins = skip_plugins if skip_plugins is not None else [plugins.AUTO_ADMIN_VIEWS]
         include_plugins = include_plugins or []
@@ -135,6 +138,11 @@ class AlchemyJsonApiViewBuilder:
         self._computed_filters[name] = AlchemyComputedFilter(name=name, filter_func=filter_func)
         return self
 
+    def add_custom_field(self, name: str,
+                         instance_callback: Callable[[Any], Any] = None) -> 'AlchemyJsonApiViewBuilder':
+        self._custom_fields[name] = CustomField(name=name, callback=instance_callback)
+        return self
+
     def _get_admin_urls(self) -> Sequence[partial]:
         admin_builder = deepcopy(self)
         admin_builder._skip_plugins = [plugins.AUTO_ADMIN_VIEWS]
@@ -159,19 +167,52 @@ class AlchemyJsonApiViewBuilder:
                                                resource_name=self._resource_name,
                                                fields=self._fields,
                                                support_relations=self._relations,
-                                               custom_field_handlers=self._custom_field_handlers)
+                                               custom_field_handlers=self._custom_field_handlers,
+                                               custom_fields=self._custom_fields)
         schema = SchemaType()
         schema_many = SchemaType(many=True)
 
-        base_query = self._base_query or self._model.objects.query()
-
         def default_permitted_objects(request, query):
             return query
 
+        def render_includes(includes, objects):
+
+            #
+            #  Go over all the keys that the user wants to include, and serialize them
+            #
+            rendered_includes = []
+            for include in includes:
+                #  Get the column on the model
+                include_on_model = getattr(self._model, include, None)
+                if include_on_model:
+                    model_property = include_on_model.property
+                    #  Get the primary key of the target relationship table
+                    target_primary_key = model_property.target.primary_key.columns.values()[0].name
+                    #  Get the local foreign key column name that connects the relationship
+                    local_column = model_property.local_columns.copy().pop()
+                    #  Get the Alchemy Model of the target relationship table
+                    target_model = model_property.mapper.class_
+                    #  Check that the item we are targeting has a JSON:API view
+                    if target_model not in _TYPE_TO_SCHEMA:
+                        raise Exception(f'No JSON:API view defined for type {target_model}')
+                    #  Fetch all the related items to be included in the result
+                    to_include = target_model.objects.query().filter(
+                        getattr(target_model, target_primary_key).in_(
+                            [getattr(item, local_column.name) for item in objects])).all()
+                    schema = _TYPE_TO_SCHEMA[target_model]
+                    target_many = schema['serializer'](many=True)
+                    #  Serialize all the included objects to JSON:API
+                    include_result = target_many.json_api_dump(to_include, schema['resource_name'],
+                                                               with_data=False)
+                    rendered_includes.extend(include_result)
+                else:
+                    raise Exception(f'Include {include} not supported on type {self._resource_name}')
+            return rendered_includes
+
         permitted_objects = self._permitted_objects or default_permitted_objects
 
         def object_get(request, identifier, *args, **kwargs) -> Tuple[Dict, int]:
-            permitted_query = permitted_objects(request, base_query)
+            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
 
             obj = None
 
@@ -186,12 +227,13 @@ class AlchemyJsonApiViewBuilder:
             if self._after_get_callback:
                 obj = self._after_get_callback(request, obj)
 
-            attributes = schema.dump(obj).data
-            attributes.pop(self._primary_key or 'id')
-            return {'id': identifier, 'type': self._resource_name, 'attributes': attributes}, HTTP_200_OK
+            result = schema.json_api_dump(obj,  self._resource_name)
+            # attributes.pop(self._primary_key or 'id', None)
+            return result, HTTP_200_OK
+            # return {'id': identifier, 'type': self._resource_name, 'attributes': attributes}, HTTP_200_OK
 
         def object_list(request, page, filters=None, includes=None, *args, **kwargs) -> Tuple[List, List, int, int]:
-            permitted_query = permitted_objects(request, base_query)
+            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
             #
             #  Apply all the filters from the URL
             #
@@ -219,69 +261,43 @@ class AlchemyJsonApiViewBuilder:
             #  Fetch the values from DB
             objects = query.all()
 
-            #
-            #  Go over all the keys that the user wants to include, and serialize them
-            #
-            rendered_includes = []
-            for include in includes:
-                #  Get the column on the model
-                include_on_model = getattr(self._model, include, None)
-                if include_on_model:
-                    model_property = include_on_model.property
-                    #  Get the primary key of the target relationship table
-                    target_primary_key = model_property.target.primary_key.columns.values()[0].name
-                    #  Get the local foreign key column name that connects the relationship
-                    local_column = model_property.local_columns.copy().pop()
-                    #  Get the Alchemy Model of the target relationship table
-                    target_model = model_property.mapper.class_
-                    #  Check that the item we are targeting has a JSON:API view
-                    if target_model not in _TYPE_TO_SCHEMA:
-                        raise Exception(f'No JSON:API view defined for type {target_model}')
-                    #  Fetch all the related items to be included in the result
-                    to_include = target_model.objects.query().filter(
-                        getattr(target_model, target_primary_key).in_(
-                            [getattr(item, local_column.name) for item in objects])).all()
-                    schema = _TYPE_TO_SCHEMA[target_model]
-                    target_many = schema['serializer'](many=True)
-                    #  Serialize all the included objects to JSON:API
-                    include_result = target_many.json_api_dump(to_include, schema['resource_name'])
-                    rendered_includes.extend(include_result.data)
-                else:
-                    raise Exception(f'Include {include} not supported on type {self._resource_name}')
+            rendered_includes = render_includes(includes, objects)
 
             if self._after_list_callback:
                 objects = self._after_list_callback(request, objects)
 
             result = schema_many.json_api_dump(objects, self._resource_name)
-            return result.data, rendered_includes, pagination.total_results, HTTP_200_OK
+            return result, rendered_includes, pagination.total_results, HTTP_200_OK
 
         def object_create(request, data, *args, **kwargs) -> Tuple[Dict, str, int]:
             if 'multipart' in request.content_type:
                 attributes = data
             else:
                 attributes = data['attributes']
-
             if self._before_create_callback:
                 attributes = self._before_create_callback(request, attributes)
 
-            unmarshal_obj = schema.load(attributes, session=schema.session)
-            obj = unmarshal_obj.data
+            try:
+                unmarshal_obj = schema.load(attributes, session=schema.db.session, unknown=INCLUDE)
+            except ValidationError as err:
+                return err.messages, '', HTTP_400_BAD_REQUEST
+
+            obj = unmarshal_obj
             obj.save()
-            obj.refresh_from_db()
 
             if self._after_create_callback:
                 self._after_create_callback(request, attributes, obj)
+                obj.refresh_from_db()
 
-            obj.refresh_from_db()
+            dumped_attributes = schema.dump(obj)
+            dumped_attributes.pop(self._primary_key or 'id', None)
 
-            return {'data': {'type': self._resource_name, 'id': obj.id, 'attributes': schema.dump(obj).data}}, obj.id, \
-                   HTTP_201_CREATED
+            return {'data': {'type': self._resource_name, 'id': obj.id,
+                             'attributes': dumped_attributes}}, obj.id, HTTP_201_CREATED
 
         def object_update(request, identifier, data, *args, **kwargs) -> Tuple[Dict, int]:
-            permitted_query = permitted_objects(request, base_query)
-
+            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
             obj = None
-
             try:
                 obj = permitted_query.filter_by(**{self._primary_key or 'id': identifier}).first()
             except StatementError:
@@ -291,28 +307,28 @@ class AlchemyJsonApiViewBuilder:
                     return {}, HTTP_404_NOT_FOUND
 
             attributes = data['attributes']
-
             if self._before_update_callback:
                 attributes = self._before_update_callback(request, attributes, obj)
+                obj.refresh_from_db()
 
             for field in self._fields:
                 if field != identifier and field in attributes:
                     setattr(obj, field, attributes[field])
-
             obj.save()
-            obj.refresh_from_db()
 
             if self._after_update_callback:
                 self._after_update_callback(request, obj)
+                obj.refresh_from_db()
 
-            obj.refresh_from_db()
+            dumped_attributes = schema.dump(obj)
+            dumped_attributes.pop(self._primary_key or 'id', None)
 
             return {'data': {'type': self._resource_name, 'id': identifier,
-                             'attributes': schema.dump(obj).data}}, HTTP_200_OK
+                             'attributes': dumped_attributes}}, HTTP_200_OK
 
         def object_delete(request, identifier, *args, **kwargs) -> Tuple[int]:
-            permitted_query = permitted_objects(request, base_query)
-            obj = permitted_query.filter(**{self._primary_key or 'id': identifier}).get()
+            permitted_query = permitted_objects(request, self._base_query or self._model.objects.query())
+            obj = permitted_query.filter_by(**{self._primary_key or 'id': identifier}).first()
             if self._before_delete_callback:
                 obj = self._before_delete_callback(request, obj)
             if obj:
